import requests
import time
import json

class Gitlab:

  DEFAULT_PAGE_SIZE = 100

  def __init__(self, token, base_url="https://gitlab.com/"):
    self.token = token
    self.base_url = base_url
    self.api_url = "api/v4/"
    self.request_headers = {
      "PRIVATE-TOKEN": self.token,
      "Content-Type": "application/json"
    }

  ### API METHODS ###

  def list_generic(self, path, page, page_size):
    response = requests.get(
      f'{self.base_url}{self.api_url}{path}?per_page={page_size}&page={page}', 
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code,
    }

  def list_groups(self):
    return self.get_paginated_list_generic('groups')

  def list_sub_groups(self, id):
    return self.get_paginated_list_generic(f'groups/{id}/subgroups')

  def list_group_projects(self, id):
    return self.get_paginated_list_generic(f'groups/{id}/projects')

  def create_group(self, path, parent_id = None):
    
    body = {
      "name": str(path.replace('-', ' ').title()),
      "path": str(path)
    }

    if parent_id is not None:
      body['parent_id'] = str(parent_id)

    response = requests.post(
      f'{self.base_url}{self.api_url}groups', 
      headers=self.request_headers, 
      data=json.dumps(body)
    )
    return {'status_code': response.status_code}

  def create_project(self, path, namespace_id):
    
    body = {
      "name": str(path.replace('-', ' ').title()),
      "path": str(path),
      "namespace_id": str(namespace_id),
    }

    response = requests.post(
      f'{self.base_url}{self.api_url}projects',
      headers=self.request_headers,
      data=json.dumps(body)
    )
    return {'status_code': response.status_code}

  def delete_project(self, project_id):

    response = requests.delete(
      f'{self.base_url}{self.api_url}projects/{project_id}', 
      headers=self.request_headers
    )
    return {'status_code': response.status_code}

  def get_project(self, project_id):

    response = requests.get(
      f'{self.base_url}{self.api_url}projects/{project_id}', 
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code,
    }

  def unprotect_branch(self, project_id, branch_name='main'):

    response = requests.delete(
      f'{self.base_url}{self.api_url}projects/{project_id}/protected_branches/{branch_name}', 
      headers=self.request_headers
    )
    return {'status_code': response.status_code}

  def create_branch(self, project_id, branch_name, branch_ref):

    response = requests.post(
      f'{self.base_url}{self.api_url}projects/{project_id}/repository/branches?branch={branch_name}&ref={branch_ref}', 
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code,
    }

  def get_branch(self, project_id, branch_name):

    response = requests.get(
      f'{self.base_url}{self.api_url}projects/{project_id}/repository/branches/{branch_name}', 
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code,
    }

  def get_paginated_list_generic(self, path, page = 1, page_size = DEFAULT_PAGE_SIZE, previous_pages = []):
    current_page = self.list_generic(path, page, page_size)['text']
    pages = current_page

    try:
      pages = previous_pages + current_page
    except TypeError:
      pass

    if len(current_page) == page_size:
      return self.get_paginated_list_generic(path, page + 1, page_size, pages)
    else:
      return pages

  def create_pages_domain(self, project_id, domain, auto_ssl_enabled = True):

    body = {
      "domain": str(domain),
      "auto_ssl_enabled": auto_ssl_enabled,
    }

    response = requests.post(
      f'{self.base_url}{self.api_url}projects/{project_id}/pages/domains',
      headers=self.request_headers,
      data=json.dumps(body)
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code
    }

  def list_pages_domain(self, project_id):

    response = requests.get(
      f'{self.base_url}{self.api_url}projects/{project_id}/pages/domains',
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code
    }

  def create_pages_domain_if_available(self, project_id, domain, auto_ssl_enabled = True):

    print('--- BUILDING PAGES ---')
    pages_create = self.create_pages_domain(project_id, domain, auto_ssl_enabled)

    if pages_create['status_code'] == 404:
      print(f"--- {pages_create['text']['message']}")
      print('--- FINISHED PAGES ---')
      return 'N/A'

    if pages_create['status_code'] == 400:
      print(f"--- Domain {domain} is already taken")
      pages_list = self.list_pages_domain(project_id)['text']

      for page in pages_list:
        if page['domain'] == domain:
          print(f"--- Verification code is {page['verification_code']}")
          print('--- FINISHED PAGES ---')
          return page['verification_code']
      
      print(f"--- Domain {domain} does not belong to this project")
      print('--- FINISHED PAGES ---')
      return 'N/A'

    print(f"--- Verification code is {pages_create['text']['verification_code']}")
    print('--- FINISHED PAGES ---')
    return pages_create['text']['verification_code']

  def get_latest_pipeline(self, project_id):

    response = requests.get(
      f'{self.base_url}{self.api_url}projects/{project_id}/pipelines/latest',
      headers=self.request_headers
    )
    return {
      'text': json.loads(response.text),
      'status_code': response.status_code
    }

  ### LOGIC CALLS ###

  def get_groups(self, parent_id = None):
    return self.list_groups() if parent_id is None else self.list_sub_groups(parent_id)

  def find_object_by_path(self, path, array_of_objs):
    for obj in array_of_objs:
      if obj['path'] == path:
        return obj

    return []

  def find_base_group_by_path(self, path, array_of_objs):
    for obj in array_of_objs:
      if obj['path'] == path and obj['full_path'] == path :
        return obj

    return []

  def build_request_chain(self, full_path):

    print(f'--- BUILDING FOR {full_path} ---')

    paths = full_path.split('/')
    parent_id = self.find_base_group_by_path(paths[0], self.get_groups())['id']

    for path in paths[1:-1]:
      parent_id = self.build_group(path, parent_id)['id']

    projs = paths[-1].split('+')
    projs_ids = []
    for path in projs:
      proj = self.build_project(path, parent_id)
      projs_ids.append(proj["id"])

    print(f'--- FINISHED BUILDING ---')
    return projs_ids

  def build_group(self, path, parent_id):
    group = self.find_object_by_path(path, self.get_groups(parent_id))

    if not group:
      print(f'--- CREATE GROUP -> {path} PARENT ID: {parent_id}')
      try:
        self.create_group(path, parent_id)
      except Exception as e:
        raise e
      group = self.find_object_by_path(path, self.get_groups(parent_id))
    else:
      print(f'--- GROUP EXISTS -> {path} WITH ID: {group["id"]} AND PARENT ID: {parent_id}')

    return group

  def build_project(self, path, parent_id):
    proj = self.find_object_by_path(path, self.list_group_projects(parent_id))

    if not proj:
      print(f'--- CREATE PRJCT -> {path} WITH PARENT ID: {parent_id}')
      try:
        self.create_project(path, parent_id)
      except Exception as e:
        raise e
      proj = self.find_object_by_path(path, self.list_group_projects(parent_id))

      print(f'--- CREATE BRANCH MAIN FOR PRJCT -> {path} WITH ID: {proj["id"]}')
      self.create_branch(proj["id"], 'main', 'main')

      print(f'--- UNPROTECTING BRANCH MAIN -> {path} {proj["id"]}')
      self.unprotect_branch(proj["id"])

    else:
      print(f'--- PRJCT EXISTS -> {path} WITH ID: {proj["id"]} AND PARENT ID: {parent_id}')

    return proj

  def wait_for_latest_pipeline(self, project_id):

    print(f"Waiting for Pipeline with PROJECT ID:{project_id} started")

    final_states = ['success', 'failed', 'canceled', 'skipped']
    status_codes = [400, 401, 403, 404, 500, 503]

    response = self.get_latest_pipeline(project_id)

    if response['status_code'] in status_codes:
      print(response['text']['message'])
      return response['text']['message']

    while(True):
      time.sleep(5)

      if response['text']['status'] in final_states:
        break

      response = self.get_latest_pipeline(project_id)
      print(f"Checking Pipeline with PROJECT ID:{project_id}...")
   
    print(f"Pipeline has finished with {response['text']['status']} state")
    return response['text']


from botocore.client import Config
import boto3
import json

class Aws:

  def __init__(self, service, access_key, secret_key, region_name="eu-west-1"):

    self.config = Config(
      read_timeout=900,
      connect_timeout=30,
      retries={"max_attempts": 0},
      tcp_keepalive=True
    )

    self.service = service
    self.access_key = access_key
    self.secret_key = secret_key
    self.region_name = region_name
    self.client = boto3.client(
      self.service, 
      aws_access_key_id=self.access_key, 
      aws_secret_access_key=self.secret_key,
      region_name = self.region_name,
      config = self.config
    )
    
  def get_user(self, user_name):
    try:
      response = self.client.get_user(UserName=user_name)
    except:
      print(f"Couldn't get user {user_name}.")
      raise

    return response

  def create_lambda(self, function_name, iam_role_arn, image_uri, subnet_ids, security_group_ids, description='AWS Lambda'):
    try:
      response = self.client.create_function(
        FunctionName=function_name,
        Description=description,
        Role=iam_role_arn,
        Code={
          'ImageUri': image_uri,
        },
        PackageType='Image',
        Timeout=900,
        MemorySize=1024,
        VpcConfig={
          'SubnetIds': subnet_ids,
          'SecurityGroupIds': security_group_ids
        },
        Publish=True
      )
      function_arn = response['FunctionArn']
      waiter = self.client.get_waiter('function_active_v2')
      waiter.wait(FunctionName=function_name)
      print(f"Created function {function_name} with ARN: {response['FunctionArn']}.")
    except:
      print(f"Couldn't create function {function_name}.")
      raise
    else:
      return function_arn

  def get_lambda(self, function_name):
    response = None
    try:
      response = self.client.get_function(FunctionName=function_name)
    except Exception as err:
      if err.response['Error']['Code'] == 'ResourceNotFoundException':
        print(f"Function {function_name} does not exist.")
      else:
        print(f"Couldn't get function {function_name}. Reason: {err.response['Error']['Code']}: {err.response['Error']['Message']}")
        raise
    return response

  def update_lambda_configuration(self, function_name, timeout=3, memory_size=128, subnet_ids=[], security_group_ids=[]):
    try:
      response = self.client.update_function_configuration(
        FunctionName=function_name,
        Timeout=timeout,
        MemorySize=memory_size,
        VpcConfig={
          'SubnetIds': subnet_ids,
          'SecurityGroupIds': security_group_ids
        }
      )
    except Exception as err:
      print(f"Couldn't get function {function_name}. Reason: {err.response['Error']['Code']}: {err.response['Error']['Message']}")
      raise
    else:
      return response

  def invoke_lambda(self, function_name, function_params, get_log=False):
    try:
      response = self.client.invoke(
        FunctionName=function_name,
        Payload=json.dumps(function_params),
        InvocationType="RequestResponse",
        LogType='Tail' if get_log else 'None'
      )
      print(f"Invoked function {function_name}")
    except:
      print(f"Couldn't invoke function {function_name}")
      raise
    return response
  
  def create_record(self, domain, record_name, record_value, record_type, hosted_zone_id):

    if record_type == 'TXT':
      record_value = f'\"{record_value}\"'

    self.client.change_resource_record_sets(
      ChangeBatch={
        'Changes': [
          {
            'Action': 'UPSERT',
            'ResourceRecordSet': {
              'Name': f'{record_name}.{domain}',
              'ResourceRecords': [
                {
                  'Value': f'{record_value}',
                },
              ],
              'TTL': 300,
              'Type': f'{record_type}',
            },
          },
        ],
      },
      HostedZoneId=hosted_zone_id,
    )

  def get_hosted_zone_id(self, name):
    response = self.client.list_hosted_zones_by_name(DNSName=name)
    if not response:
      return 'N/A'
    elif response['HostedZones'][0]['Name'] == f'{name}.':
      return response['HostedZones'][0]['Id']
    return 'N/A'


from vanilla_violin import Gitlab
from ddt import ddt, data, unpack

import unittest

@ddt
class TestFindGroupByPath(unittest.TestCase):

  GROUPS = [
    {
      "id":56789,
      "path":"badjumbas"
    },
    {
      "id":12345,
      "path":"massusso"
    },
    {
      "id":11111,
      "path":"alioli"
    }
  ]

  @data(('massusso', {'id': 12345, 'path': 'massusso'}), ('nao-existo-lel', []))
  @unpack
  def test_group_found(self, value, expectation):
    gitlab = Gitlab('')
    group = gitlab.find_object_by_path(value, self.GROUPS)
    self.assertEqual(group, expectation)


from setuptools import setup
setup(
  name='vanilla-violin',
  version='0.0.9',
  packages=[
    'vanilla_violin',
    'vanilla_violin.gitlab',
    'vanilla_violin.aws',
  ],
  entry_points={
    'console_scripts': [
      'co=cli.__main__:cli',
    ]
  },
  install_requires=[
    "Click==8.1.3",
    "ddt==1.6.0",
    "boto3==1.26.56",
    "requests==2.28.1",
    "pytest==7.2.0",
    "PyYAML==6.0",
  ],
)
